import msgpack

from plantstar_shared.errors import SocketConnectionError


# Converts bytes into an object using msgpack, with a parameter for strict_map_key
def convert_bytes_to_object(object_as_bytes, strict_map_key=False):
    if not object_as_bytes:
        raise SocketConnectionError
    
    output_object = msgpack.unpackb(object_as_bytes, strict_map_key=strict_map_key)
    return output_object



class HmiKeyConstants:
    APU_IP_ADDRESS = "apu_ip_address"
    HMI_HOSTNAME = "hmi_hostname"
    HMI_HOMEPAGE = "hmi_homepage"
    HMI_IP_ADDRESS_TYPE = "hmi_ip_address_type"
    HMI_IP_ADDRESS = "hmi_ip_address"
    HMI_NETMASK = "hmi_netmask"
    HMI_DEFAULT_GATEWAY = "hmi_default_gateway"
    HMI_DNS = "hmi_dns"
    HMI_LAN_TYPE = "hmi_lan_type"
    HMI_WIRELESS_SSID = "hmi_wireless_ssid"
    HMI_WIRELESS_PASSPHRASE = "hmi_wireless_passphrase"
    HMI_MACHINE_TYPE = "hmi_machine_type"


class HmiMachineTypeConstants:
    RPI4_HMI = "rpi4hmi"
    ALL_IN_ONE_HMI = "allinonehmi"



import datetime
import time
import array
from itertools import chain, repeat, islice

from plantstar_shared.global_definitions import utc_now

TEN_MILLISECONDS = .01


class MockRawDataProcessor:
    def __init__(
        self, cycle_time_in_milliseconds=30000, number_of_digital_inputs=200, number_of_digital_outputs=200, number_of_analog_inputs=200, debounce_raw_buffer_size=255,
        analog_value_min=50, analog_value_max=2, analog_value_step=0.5
    ):
        self.next_value_swap_in_seconds_timedelta = datetime.timedelta(milliseconds=cycle_time_in_milliseconds / 2)
        self.next_value_swap_datetime = utc_now() + self.next_value_swap_in_seconds_timedelta

        self.number_of_digital_inputs = number_of_digital_inputs
        self.number_of_digital_outputs = number_of_digital_outputs
        self.number_of_analog_inputs = number_of_analog_inputs
        self.debounce_raw_buffer_size = debounce_raw_buffer_size
        self.analog_value_min = analog_value_min
        self.analog_value_max = analog_value_max
        self.analog_value_step = analog_value_step

        self.last_digital_input_debounce_array = array.array("B", pad([0 for _ in range(self.number_of_digital_inputs)], self.debounce_raw_buffer_size, 0))
        self.last_digital_output_array = array.array("B", pad([0 for _ in range(self.number_of_digital_outputs)], self.debounce_raw_buffer_size, 0))
        self.last_analog_filter_array = array.array("B", pad([0 for _ in range(self.number_of_analog_inputs)], self.debounce_raw_buffer_size, 0))
        self.last_analog_gain_array = array.array("B", pad([0 for _ in range(self.number_of_analog_inputs)], self.debounce_raw_buffer_size, 0))

        self.state_on_or_off_value = False
        self.current_analog_value = analog_value_max
        self.current_analog_step_direction = -1
        self.cold_end_value = 20

    def upload_new_driver_settings(self, data_type_string, data_buffer):
        data_buffer_array = array.array("B", list(pad(data_buffer, self.debounce_raw_buffer_size, 0)))

        if data_type_string == "digital_outputs":
            if self.last_digital_output_array != data_buffer_array:
                self.last_digital_output_array = data_buffer_array
        elif data_type_string == "analog_gains":
            if self.last_analog_gain_array != data_buffer_array:
                self.last_analog_gain_array = data_buffer_array
        elif data_type_string == "analog_filters":
            if self.last_analog_filter_array != data_buffer_array:
                self.last_analog_filter_array = data_buffer_array
        elif data_type_string == "digital_input_debounce":
            if self.last_digital_input_debounce_array != data_buffer_array:
                self.last_digital_input_debounce_array = data_buffer_array
        else:
            if isinstance(data_type_string, str):
                raise ValueError(
                    f"An Incorrect value of {data_type_string} was passed. This function expects \"digital_outputs\", \"analog_gains\", or \"analog_filters\"."
                )
            else:
                raise TypeError(
                    f"Failed to pass a recognizable type of data. {data_type_string} was passed in with a {type(data_type_string)} data type. This function expects a string "
                    f"of \"digital_outputs\", \"analog_gains\", or \"analog_filters\"."
                )

    def get_input_signal_dictionaries(self):
        time.sleep(TEN_MILLISECONDS)

        the_now = utc_now()

        if self.next_value_swap_datetime <= the_now:
            self.state_on_or_off_value = not self.state_on_or_off_value
            self.next_value_swap_datetime = the_now + self.next_value_swap_in_seconds_timedelta

            if self.state_on_or_off_value:
                self.current_analog_value = self.analog_value_max
                self.current_analog_step_direction = -1

        return_input_signal_dictionary_list = [
            {
                "analogs": [self.current_analog_value for _ in range(self.debounce_raw_buffer_size)],
                "cold_ends": [self.cold_end_value for _ in range(self.debounce_raw_buffer_size)],
                "digitals": [self.state_on_or_off_value for _ in range(self.debounce_raw_buffer_size)],
                "event_unix_timestamp": the_now.timestamp()
            }
        ]

        self.step_current_analog_value()

        return return_input_signal_dictionary_list

    def get_driver_version_string(self):
        return "Mock"

    def get_version_string(self):
        return "Mock"

    def step_current_analog_value(self):
        if self.current_analog_step_direction == -1:
            step_min_max_function = max
            ceiling_or_floor_value = self.analog_value_min
        else:
            step_min_max_function = min
            ceiling_or_floor_value = self.analog_value_max

        new_analog_value = step_min_max_function(self.current_analog_value + self.analog_value_step * self.current_analog_step_direction, ceiling_or_floor_value)

        if new_analog_value == ceiling_or_floor_value:
            self.current_analog_step_direction *= -1

        self.current_analog_value = new_analog_value


def pad(iterable, size, padding=None):
    return islice(pad_infinite(iterable, padding), size)


def pad_infinite(iterable, padding=None):
    return chain(iterable, repeat(padding))



import msgpack

# Converts the output dictionary into bytes using msgpack
def convert_object_to_bytes(input_object):
    input_object_as_bytes = msgpack.packb(input_object)
    return input_object_as_bytes



import struct

from plantstar_shared.convert_bytes_to_object import convert_bytes_to_object
from plantstar_shared.convert_object_to_bytes import convert_object_to_bytes
from plantstar_shared.errors import SocketConnectionError, SysconProgrammingError

SIZE_OF_UNSIGNED_INT_FOR_HUSKY = 2  # size of the integer value that precedes data coming from Husky OIs
SIZE_OF_UNSIGNED_INT_STRUCT = 4  # size of the integer value that precedes data coming from most sockets (DeviceController, DCM, etc)


def read_size_value_from_socket(*, remote_socket, is_big_endian, number_of_bytes_for_size_prefix, size_value_includes_size_prefix_bytes=False):
    size_bytes = remote_socket.recv(number_of_bytes_for_size_prefix)

    if not size_bytes:
        raise SocketConnectionError

    if number_of_bytes_for_size_prefix == SIZE_OF_UNSIGNED_INT_FOR_HUSKY:
        size_for_message = int.from_bytes(size_bytes, 'big') if is_big_endian else int.from_bytes(size_bytes, 'little')
    elif number_of_bytes_for_size_prefix == SIZE_OF_UNSIGNED_INT_STRUCT:
        number_format = ">I" if is_big_endian else "<I"

        try:
            size_for_message = struct.unpack(number_format, size_bytes)[0]
        except struct.error as error:
            raise SocketConnectionError
    else:
        raise SysconProgrammingError("A size prefix was set to something other than 2 or 4")

    if size_value_includes_size_prefix_bytes:
        size_for_message = size_for_message - number_of_bytes_for_size_prefix

    return size_for_message, size_bytes


def get_bytes_from_socket(
    *, remote_socket, number_of_bytes_to_read=None, is_big_endian=True, number_of_bytes_for_size_prefix=0, size_value_includes_size_prefix_bytes=False,
    size_value_should_be_included_in_final_message=False
):
    """Obtains a number of bytes from a provided socket, with options for different configurations and situations.
    Returns byte sequence, and the size value that was obtained from the prefix bytes.

    Keywords / Example Configurations:
        remote_socket -- the connection that bytes will be read from

        number_of_bytes_to_read -- the number of bytes to read on the socket, if this is known, this is the only optional parameter needed (default: 0)

        is_big_endian -- bool: what format string should be used for unpacking the size value (default: True)

        number_of_bytes_for_size_prefix -- the number of bytes to read to determine the total size of the message to pass (default: 0)

        size_value_includes_size_prefix_bytes -- bool: if size value includes the size_prefix_bytes based on the systems implementation

        size_value_should_be_included_in_final_message  -- bool: if size value should be prepended to the byte sequence, likely only Husky logic wants this (default: False)

        -----

        get_bytes_from_socket(remote_socket=conn, number_of_bytes_to_read=100) - likely would not be called directly, in most cases

        get_bytes_from_socket(remote_socket=conn, is_big_endian=True, number_of_bytes_for_size_prefix=4) - reads 4 bytes to get the size, then reads that amount (DCM socket, etc)

        get_bytes_from_socket(remote_socket=conn, is_big_endian=False, number_of_bytes_for_size_prefix=2, should_remove_prefix_size_from_read=True) - reads 2 bytes to get the size,
        then reads that calculated amount minus 2 (Husky socket, etc)

    """

    packets = []

    if not number_of_bytes_to_read:
        if number_of_bytes_for_size_prefix:
            number_of_bytes_to_read, size_bytes = read_size_value_from_socket(
                remote_socket=remote_socket, is_big_endian=is_big_endian, number_of_bytes_for_size_prefix=number_of_bytes_for_size_prefix,
                size_value_includes_size_prefix_bytes=size_value_includes_size_prefix_bytes
            )

            if not number_of_bytes_to_read:
                raise SocketConnectionError

            if size_value_should_be_included_in_final_message:
                packets.append(size_bytes)

        else:
            raise SysconProgrammingError("get_bytes_from_socket was called with no parameters for number_of_bytes_to_read or number_of_bytes_for_size_prefix")

    bytes_received = 0

    # While loop that will stream in data if the full request is not available yet
    while bytes_received < number_of_bytes_to_read:
        buffer_size = number_of_bytes_to_read - bytes_received
        packet = remote_socket.recv(buffer_size)

        if not packet:
            raise SocketConnectionError

        packets.append(packet)
        bytes_received = bytes_received + len(packet)

    data = b''.join(packets)
    return data, number_of_bytes_to_read


def send_message_on_socket(*, remote_socket, dumpsable_object):
    message_encoded_as_bytes = convert_object_to_bytes(dumpsable_object)
    send_encoded_message_on_socket(remote_socket=remote_socket, encoded_message=message_encoded_as_bytes)


def send_encoded_message_on_socket(*, remote_socket, encoded_message):
    size_of_string = len(encoded_message)
    encoded_message_with_size = struct.pack('>I', size_of_string) + encoded_message

    bytes_sent = remote_socket.send(encoded_message_with_size)

    while bytes_sent < size_of_string + SIZE_OF_UNSIGNED_INT_STRUCT:
        remaining_encoded_message_to_send = encoded_message_with_size[bytes_sent:]
        new_bytes_sent = remote_socket.send(remaining_encoded_message_to_send)
        bytes_sent += new_bytes_sent


def get_object_from_socket(
    *, remote_socket, number_of_bytes_to_read=None, is_big_endian=True, number_of_bytes_for_size_prefix=SIZE_OF_UNSIGNED_INT_STRUCT,
    size_value_includes_size_prefix_bytes=False
):
    """Function that is used between the APU and DCM to send dictionaries. Returns a tuple of the object and its size in bytes."""
    object_from_interface_as_bytes, size_of_object_in_bytes = get_bytes_from_socket(
        remote_socket=remote_socket, number_of_bytes_to_read=number_of_bytes_to_read, is_big_endian=is_big_endian, number_of_bytes_for_size_prefix=number_of_bytes_for_size_prefix,
        size_value_includes_size_prefix_bytes=size_value_includes_size_prefix_bytes
    )

    if not object_from_interface_as_bytes:
        raise SocketConnectionError

    object_from_interface = convert_bytes_to_object(object_from_interface_as_bytes)
    return object_from_interface, size_of_object_in_bytes
